import re
from typing import Annotated
from pydantic import BeforeValidator

# Regex pattern to match any characters that aren't alphanumeric, underscore, or hyphen
_regex = re.compile(r"[^a-zA-Z0-9_-]+")


def path_safe_name(name: str) -> str:
    """
    Convert a string to a filesystem-safe format by replacing unsafe characters with underscores.

    Args:
        name: The string to be converted to a path-safe format

    Returns:
        A string with all characters not matching [a-zA-Z0-9_-] replaced with underscores

    Example:
        >>> path_safe_name("Hello World!")
        'Hello_World_'
        >>> path_safe_name("file:name.txt")
        'file_name_txt'
    """
    return _regex.sub("_", name)


# Custom type that automatically converts strings to path-safe format
PathSafeStr = Annotated[str, BeforeValidator(path_safe_name)]
"""
A custom Pydantic type that automatically sanitizes strings for safe use in file paths.
When a field is annotated with this type, any unsafe characters will be replaced with underscores.

Example:
    >>> from pydantic import BaseModel
    >>> class FileModel(BaseModel):
    ...     filename: PathSafeStr
    ...
    >>> file = FileModel(filename="my file:name.txt")
    >>> file.filename
    'my_file_name_txt'
"""



from __future__ import annotations

import importlib
from typing import TypeVar, TYPE_CHECKING

if TYPE_CHECKING:
    from bfabric.entities.core.entity import Entity

E = TypeVar("E", bound="Entity")


def import_entity(entity_class_name: str) -> type[E]:
    """Returns the entity class from a string, e.g. for `"Workunit"` returns the Workunit class."""
    return importlib.import_module(f"bfabric.entities.{entity_class_name.lower()}").__dict__[
        entity_class_name.capitalize()
    ]



from __future__ import annotations

import re
from typing import Any

from loguru import logger
from zeep.helpers import serialize_object

from bfabric.engine.response_format_suds import suds_asdict_recursive
from bfabric.errors import BfabricRequestError
from bfabric.results.result_container import ResultContainer


class ResponseDelete(ResultContainer):
    """A response to a `delete` request.

    Any errors will be contained in errors, whereas the result will be a list of dictionaries with `id` of only the
    successfully deleted entities.
    """

    __msg_success_pattern = re.compile(r"\w+ (\d+) removed successfully\.")
    __report_key = "deletionreport"

    @classmethod
    def from_empty_request(cls) -> ResponseDelete:
        """Creates an empty `ResponseDelete`."""
        logger.warning("Attempted to delete an empty list, no request will be performed.")
        return cls(results=[], errors=[], total_pages_api=0)

    @classmethod
    def from_suds(cls, suds_response: Any, endpoint: str) -> ResponseDelete:
        """Creates a `ResponseDelete` from a SUDS response."""
        result_parsed = [suds_asdict_recursive(entry, convert_types=True) for entry in suds_response[endpoint]]
        results, errors = cls.__convert_parsed_response(result_parsed)
        return cls(results=results, errors=errors, total_pages_api=None)

    @classmethod
    def from_zeep(cls, zeep_response: Any, endpoint: str) -> ResponseDelete:
        """Creates a `ResponseDelete` from a ZEEP response."""
        result_parsed = [dict(serialize_object(result, target_cls=dict)) for result in zeep_response[endpoint]]
        results, errors = cls.__convert_parsed_response(result_parsed)
        return cls(results=results, errors=errors, total_pages_api=None)

    @classmethod
    def __convert_parsed_response(
        cls, result_parsed: list[dict[str, Any]]
    ) -> tuple[list[dict[str, Any]], list[BfabricRequestError]]:
        """Converts the response to a `delete` request."""
        results = []
        errors = []
        for result in result_parsed:
            if "errorreport" in result and result["errorreport"]:
                errors.append(BfabricRequestError(result["errorreport"]))
                continue
            report_msg = result[cls.__report_key]
            match_success = cls.__msg_success_pattern.match(report_msg)
            if match_success:
                results.append({"id": int(match_success.group(1)), cls.__report_key: report_msg})
            else:
                errors.append(BfabricRequestError(report_msg))
        return results, errors



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Literal

from pydantic import BaseModel

from bfabric.config import BfabricClientConfig, BfabricAuth  # noqa
from bfabric.config.config_file import read_config_file


class ConfigData(BaseModel):
    client: BfabricClientConfig
    auth: BfabricAuth | None

    def with_auth(self, auth: BfabricAuth | None) -> ConfigData:
        """Returns a shallow copy of self with the auth field set to the specified value."""
        return ConfigData(client=self.client, auth=auth)


def _read_config_file(config_path: Path | str, force_config_env: str | None) -> ConfigData:
    """Reads the config file and returns the config data."""
    config_file_path = Path(config_path).expanduser()
    if not config_file_path.is_file():
        msg = f"No explicit config provided, and no config file found at {config_file_path}"
        raise OSError(msg)

    config, auth = read_config_file(
        config_path=config_file_path, config_env=force_config_env or os.environ.get("BFABRICPY_CONFIG_ENV")
    )
    return ConfigData(client=config, auth=auth)


def load_config_data(
    config_file_path: Path | str,
    config_file_env: str | Literal["default"] | None,
    include_auth: bool,
) -> ConfigData:
    """Loads the configuration data."""
    if "BFABRICPY_CONFIG_OVERRIDE" in os.environ:
        config_data = ConfigData.model_validate_json(os.environ["BFABRICPY_CONFIG_OVERRIDE"])
    elif config_file_env is not None:
        config_file_env = os.environ.get("BFABRICPY_CONFIG_ENV") if config_file_env == "default" else config_file_env
        config_data = _read_config_file(config_path=config_file_path, force_config_env=config_file_env)
    else:
        msg = "No configuration was found and config_file_env is set to None."
        raise ValueError(msg)
    return config_data if include_auth else config_data.with_auth(None)


def export_config_data(config_data: ConfigData) -> str:
    """Export the config data as a JSON string."""
    auth_data = config_data.auth.model_dump() if config_data.auth else None
    if auth_data is not None:
        auth_data["password"] = auth_data["password"].get_secret_value()
    data = {"client": config_data.client.model_dump(mode="json", round_trip=True), "auth": auth_data}
    return json.dumps(data)



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from pydantic import TypeAdapter

from bfabric.entities.core.entity import Entity
from bfabric.entities.core.has_one import HasOne
from bfabric.utils.path_safe_name import PathSafeStr

if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric.entities.executable import Executable
    from bfabric.entities.storage import Storage


class Application(Entity):
    ENDPOINT = "application"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    storage: HasOne[Storage] = HasOne("Storage", bfabric_field="storage")
    executable: HasOne[Executable] = HasOne("Executable", bfabric_field="executable")

    @property
    def technology_folder_name(self) -> PathSafeStr:
        """Returns the technology which is used e.g. for output registration."""
        technology = self.data_dict["technology"]
        if isinstance(technology, list):
            # TODO certainly this can be improved, also in the future it may always be a list (right now it is not
            #      rolled out yet)
            technology = sorted(technology)[0]
        return TypeAdapter(PathSafeStr).validate_python(technology)
